@@ -37,6 +37,7 @@ use File::Temp qw/ tempfile tempdir /;
 use Cwd qw(abs_path getcwd);
 use autouse 'Carp' => qw(longmess confess);
 use Scalar::Util 'reftype';
+use FindBin;
 
 our @ISA = qw(Exporter);
 #
@@ -60,6 +61,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 				       disable_script
 		                       numeric_value
 		                       numeric_value1
+				       normalize_hex
 		                       hex_value
 		                       in_hex
 		                       in_hex2
@@ -126,6 +128,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 				       $debug
 				       %config
 				       %globals
+				       %config_files
 
 		                       F_IPV4
 		                       F_IPV6
@@ -136,7 +139,7 @@ our %EXPORT_TAGS = ( internal => [ qw( create_temp_script
 
 Exporter::export_ok_tags('internal');
 
-our $VERSION = '4.4_17';
+our $VERSION = '4.4_19';
 
 #
 # describe the current command, it's present progressive, and it's completion.
@@ -262,6 +265,50 @@ our %capdesc = ( NAT_ENABLED     => 'NAT',
 		 CAPVERSION      => 'Capability Version',
 		 KERNELVERSION   => 'Kernel Version',
 	       );
+
+our %config_files = ( #accounting      => 1,
+		      actions          => 1,
+		      blacklist        => 1,
+		      clear            => 1,
+		      ecn              => 1,
+		      findgw           => 1,
+		      hosts            => 1,
+		      init             => 1,
+		      initdone         => 1,
+		      interfaces       => 1,
+		      isusable         => 1,
+		      maclist          => 1,
+		      masq             => 1,
+		      nat              => 1,
+		      netmap           => 1,
+		      notrack          => 1,
+		      params           => 1,
+		      policy           => 1,
+		      providers        => 1,
+		      proxyarp         => 1,
+		      refresh          => 1,
+		      refreshed        => 1,
+		      restored         => 1,
+		      route_rules      => 1,
+		      routes           => 1,
+		      routestopped     => 1,
+		      rules            => 1,
+		      scfilter         => 1,
+		      secmarks         => 1,
+		      start            => 1,
+		      started          => 1,
+		      stop             => 1,
+		      stopped          => 1,
+		      tcclasses        => 1,
+		      tcclear          => 1,
+		      tcdevices        => 1,
+		      tcfilters        => 1,
+		      tcinterfaces     => 1,
+		      tcpri            => 1,
+		      tcrules          => 1,
+		      tos              => 1,
+		      tunnels          => 1,
+		      zones            => 1 );
 #
 # Directories to search for configuration files
 #
@@ -365,7 +412,7 @@ sub initialize( $ ) {
 		    EXPORT     => 0,
 		    STATEMATCH => '-m state --state',
 		    UNTRACKED  => 0,
-		    VERSION    => "4.4.17",
+		    VERSION    => "4.4.19.3",
 		    CAPVERSION => 40417 ,
 		  );
     #
@@ -773,6 +820,16 @@ sub hex_value( $ ) {
     use warnings;
 }
 
+#
+# Strip off superfluous leading zeros from a hex number
+#
+sub normalize_hex( $ ) {
+    my $val = lc shift;
+
+    $val =~ s/^0// while $val =~ /^0/ && length $val > 1;
+    $val;
+}
+
 #
 # Return the argument expressed in Hex
 #
@@ -1325,7 +1382,6 @@ sub pop_include() {
 # This function is normally called below in read_a_line() when EOF is reached. Clients of the
 # module may also call the function to close the file before EOF
 #
-
 sub close_file() {
     if ( $currentfile ) {
 	my $result = close $currentfile;
@@ -2022,9 +2078,10 @@ sub default_log_level( $$ ) {
 #
 sub check_trivalue( $$ ) {
     my ( $var, $default) = @_;
-    my $val = lc( $config{$var} || '' );
+    my $val = $config{$var};
 
     if ( defined $val ) {
+	$val = lc $val;
 	if ( $val eq 'yes' || $val eq 'on' ) {
 	    $config{$var} = 'on';
 	} elsif ( $val eq 'no' || $val eq 'off' ) {
@@ -2037,7 +2094,7 @@ sub check_trivalue( $$ ) {
 	    fatal_error "Invalid value ($val) for $var";
 	}
     } else {
-	$config{var} = $default
+	$config{$var} = $default
     }
 }
 
@@ -2860,7 +2917,7 @@ sub get_params() {
     if ( -f $fn ) {
 	progress_message2 "Processing $fn ...";
 
-	my $command = "$globals{SHAREDIRPL}/getparams $fn " . join( ':', @config_path );
+	my $command = "$FindBin::Bin/getparams $fn " . join( ':', @config_path );
 	#
 	# getparams silently sources the params file under 'set -a', then executes 'export -p'
 	#
@@ -2896,30 +2953,33 @@ sub get_params() {
 			s/"$//;
 			$params{$variable} .= $_;
 		    } else {
+			chomp;
 			warning_message "Param line ($_) ignored" unless $bug++;
 		    }
 		}	
 	    }
-	} elsif ( $params[0] =~ /^export (.*?)="/ ) {
+	} elsif ( $params[0] =~ /^export .*?="/ || $params[0] =~ /^export [^\s=]+\s*$/ ) {
 	    #
 	    # getparams interpreted by older (e.g., RHEL 5) Bash
 	    #
 	    # - Variable names preceded by 'export '
 	    # - Variable values are delimited by double quotes
 	    # - Embedded single quotes are escaped with '\'
+	    # - Valueless variables ( e.g., 'export foo') are supported
 	    #
 	    for ( @params ) {
 		if ( /^export (.*?)="(.*[^\\])"$/ ) {
 		    $params{$1} = $2 unless $1 eq '_';
 		} elsif ( /^export (.*?)="(.*)$/ ) {
 		    $params{$variable=$1} = $2 eq '"' ? '' : "${2}\n";
-		} elsif ( /^export (.*)\s+$/ || /^export (.*)=""$/ ) {
+		} elsif ( /^export ([^\s=]+)\s*$/ || /^export (.*)=""$/ ) {
 		    $params{$1} = '';
 		} else {
 		    if ($variable) {
 			s/"$//;
 			$params{$variable} .= $_;
 		    } else {
+			chomp;
 			warning_message "Param line ($_) ignored" unless $bug++;
 		    }
 		}	
@@ -2944,6 +3004,7 @@ sub get_params() {
 			s/'$//;
 			$params{$variable} .= $_;
 		    } else {
+			chomp;
 			warning_message "Param line ($_) ignored" unless $bug++;
 		    }				
 		}
@@ -2954,7 +3015,7 @@ sub get_params() {
 	    print "PARAMS:\n";
 	    my $value;
 	    while ( ($variable, $value ) = each %params ) {
-		print "   $variable='$value'\n";
+		print "   $variable='$value'\n" unless $compiler_params{$variable};
 	    }
 	}
     }
@@ -2990,7 +3051,7 @@ sub export_params() {
 
 	emit "#\n# From the params file\n#" unless $count++;
 
-	if ( $value =~ /\s/ ) {
+	if ( $value =~ /[\s()[]/ ) {
 	    emit "$param='$value'";
 	} else {
 	    emit "$param=$value";
@@ -3029,11 +3090,12 @@ sub get_configuration( $ ) {
     #
     # get_capabilities requires that the true settings of these options be established
     #
-    default 'MODULE_PREFIX', 'o gz ko o.gz ko.gz';
+    default 'MODULE_PREFIX', 'ko ko.gz o o.gz gz';
     default_yes_no 'LOAD_HELPERS_ONLY'          , '';
 
     get_capabilities( $export );
 
+
     $globals{STATEMATCH} = '-m conntrack --ctstate' if have_capability 'CONNTRACK_MATCH';
 
     if ( my $rate = $config{LOGLIMIT} ) {
@@ -3333,7 +3395,7 @@ sub get_configuration( $ ) {
 	my @priomap = split ' ',$val;
 	fatal_error "Invalid TC_PRIOMAP ($val)" unless @priomap == 16;
 	for ( @priomap ) {
-	    fatal_error "Invalid TC_PRIOMAP entry ($_)" unless /[1-3]/;
+	    fatal_error "Invalid TC_PRIOMAP entry ($_)" unless /^[1-3]$/;
 	    $_--;
 	}
 
