@@ -20,10 +20,12 @@
 #       along with this program; if not, write to the Free Software
 #       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 #
-#   This module contains process_rule() and it's associated helpers for handling 
-#   Actions and Macros.
+#   This module handles policies and rules. It contains:
 #
-#   This module combines the former Rules and Actions modules.
+#       process_policies() and it's associated helpers.
+#       process_rules() and it's associated helpers for handling Actions and Macros.
+#
+#   This module combines the former Policy, Rules and Actions modules.
 #
 package Shorewall::Rules;
 require Exporter;
@@ -31,20 +33,38 @@ use Shorewall::Config qw(:DEFAULT :internal);
 use Shorewall::Zones;
 use Shorewall::Chains qw(:DEFAULT :internal);
 use Shorewall::IPAddrs;
-use Shorewall::Policy;
 use Scalar::Util 'reftype';
 
 use strict;
 
 our @ISA = qw(Exporter);
 our @EXPORT = qw(
+		  process_policies
+		  apply_policy_rules
+		  complete_standard_chain
+		  setup_syn_flood_chains
+		  save_policies
+		  ensure_rules_chain
+		  optimize_policy_chains
 		  process_actions1
 		  process_actions2
 		  process_rules
 	       );
 
 our @EXPORT_OK = qw( initialize );
-our $VERSION = '4.4_17';
+our $VERSION = '4.4_19';
+#
+# Globals are documented in the initialize() function
+#
+our %sections;
+
+our $section;
+
+our @policy_chains;
+
+our %policy_actions;
+
+our %default_actions;
 
 our %macros;
 
@@ -55,7 +75,7 @@ our @builtins;
 #
 # Commands that can be embedded in a basic rule and how many total tokens on the line (0 => unlimited).
 #
-our $rule_commands = { COMMENT => 0, FORMAT => 2 };
+our $rule_commands = { COMMENT => 0, FORMAT => 2, SECTION => 2 };
 
 use constant { MAX_MACRO_NEST_LEVEL => 5 };
 
@@ -86,11 +106,57 @@ our %usedactions;
 #
 sub initialize( $ ) {
     $family            = shift;
+    #
+    # Chains created as a result of entries in the policy file
+    @policy_chains  = ();
+    #
+    # Default Actions for policies
+    #
+    %policy_actions = ();
+    #
+    # This is updated from the *_DEFAULT settings in shorewall.conf. Those settings were stored
+    # in the %config hash when shorewall[6].conf was processed.
+    #
+    %default_actions  = ( DROP     => 'none' ,
+		 	  REJECT   => 'none' ,
+			  ACCEPT   => 'none' ,
+			  QUEUE    => 'none' ,
+			  NFQUEUE  => 'none' ,
+			);
+    #
+    # These are set to 1 as sections are encountered.
+    #
+    %sections = ( ESTABLISHED => 0,
+		  RELATED     => 0,
+		  NEW         => 0
+		  );
+    #
+    # Current rules file section.
+    #
+    $section  = '';
+    #
+    # Macro=><macro file> mapping
+    #
     %macros            = ();
+    #
+    # Stack of nested action calls while parsing action.* files.
+    #
     @actionstack       = ();
+    #
+    # This hash provides keyed access to @actionstack
+    #
     %active            = ();
+    #
+    # Self-explainatory
+    #
     $macro_nest_level  = 0;
+    #
+    # All builtin actions plus those mentioned in /etc/shorewall[6]/actions and /usr/share/shorewall[6]/actions
+    #
     %actions           = ();
+    #
+    # Action variants actually used. Key is <action>:<loglevel>:<tag>:<params>; value is corresponding chain name
+    #
     %usedactions       = ();
 
     if ( $family == F_IPV4 ) {
@@ -100,6 +166,610 @@ sub initialize( $ ) {
     }
 }
 
+###############################################################################
+# Functions moved from the former Policy Module
+###############################################################################
+#
+# Split the passed target into the basic target and parameter (previously duplicated in this file)
+#
+sub get_target_param( $ ) {
+    my ( $target, $param ) = split '/', $_[0];
+
+    unless ( defined $param ) {
+	( $target, $param ) = ( $1, $2 ) if $target =~ /^(.*?)[(](.*)[)]$/;
+    }
+
+    ( $target, $param );
+}
+
+#
+# Convert a chain into a policy chain.
+#
+sub convert_to_policy_chain($$$$$)
+{
+    my ($chainref, $source, $dest, $policy, $provisional ) = @_;
+
+    $chainref->{is_policy}   = 1;
+    $chainref->{policy}      = $policy;
+    $chainref->{provisional} = $provisional;
+    $chainref->{policychain} = $chainref->{name};
+    $chainref->{policypair}  = [ $source, $dest ];
+}
+
+#
+# Create a new policy chain and return a reference to it.
+#
+sub new_policy_chain($$$$)
+{
+    my ($source, $dest, $policy, $provisional) = @_;
+
+    my $chainref = new_chain( 'filter', rules_chain( ${source}, ${dest} ) );
+
+    convert_to_policy_chain( $chainref, $source, $dest, $policy, $provisional );
+
+    $chainref;
+}
+
+#
+# Set the passed chain's policychain and policy to the passed values.
+#
+sub set_policy_chain($$$$$)
+{
+    my ($source, $dest, $chain1, $chainref, $policy ) = @_;
+
+    my $chainref1 = $filter_table->{$chain1};
+
+    $chainref1 = new_chain 'filter', $chain1 unless $chainref1;
+
+    unless ( $chainref1->{policychain} ) {
+	if ( $config{EXPAND_POLICIES} ) {
+	    #
+	    # We convert the canonical chain into a policy chain, using the settings of the
+	    # passed policy chain.
+	    #
+	    $chainref1->{policychain} = $chain1;
+	    $chainref1->{loglevel}    = $chainref->{loglevel} if defined $chainref->{loglevel};
+
+	    if ( defined $chainref->{synparams} ) {
+		$chainref1->{synparams}   = $chainref->{synparams};
+		$chainref1->{synchain}    = $chainref->{synchain};
+	    }
+
+	    $chainref1->{default}     = $chainref->{default} if defined $chainref->{default};
+	    $chainref1->{is_policy}   = 1;
+	    push @policy_chains, $chainref1;
+	} else {
+	    $chainref1->{policychain} = $chainref->{name};
+	}
+
+	$chainref1->{policy} = $policy;
+	$chainref1->{policypair} = [ $source, $dest ];
+    }
+}
+
+#
+# Process the policy file
+#
+use constant { PROVISIONAL => 1 };
+
+sub add_or_modify_policy_chain( $$ ) {
+    my ( $zone, $zone1 ) = @_;
+    my $chain    = rules_chain( ${zone}, ${zone1} );
+    my $chainref = $filter_table->{$chain};
+
+    if ( $chainref ) {
+	unless( $chainref->{is_policy} ) {
+	    convert_to_policy_chain( $chainref, $zone, $zone1, 'CONTINUE', PROVISIONAL );
+	    push @policy_chains, $chainref;
+	}
+    } else {
+	push @policy_chains, ( new_policy_chain $zone, $zone1, 'CONTINUE', PROVISIONAL );
+    }
+}
+
+sub print_policy($$$$) {
+    my ( $source, $dest, $policy , $chain ) = @_;
+    unless ( ( $source eq 'all' ) || ( $dest eq 'all' ) ) {
+	if ( $policy eq 'CONTINUE' ) {
+	    my ( $sourceref, $destref ) = ( find_zone($source) ,find_zone( $dest ) );
+	    warning_message "CONTINUE policy between two un-nested zones ($source, $dest)" if ! ( @{$sourceref->{parents}} || @{$destref->{parents}} );
+	}
+	progress_message_nocompress "   Policy for $source to $dest is $policy using chain $chain" unless $source eq $dest;
+    }
+}
+
+#
+# Add the passed action to %policy_actions
+#
+sub use_policy_action( $ ) {
+    my $action = shift;
+
+    $policy_actions{$action} = 1;
+}
+
+#
+# Process an entry in the policy file.
+#
+sub process_a_policy() {
+
+    our %validpolicies;
+    our @zonelist;
+
+    my ( $client, $server, $originalpolicy, $loglevel, $synparams, $connlimit ) = split_line 3, 6, 'policy file';
+
+    $loglevel  = '' if $loglevel  eq '-';
+    $synparams = '' if $synparams eq '-';
+    $connlimit = '' if $connlimit eq '-';
+
+    my $clientwild = ( "\L$client" eq 'all' );
+
+    fatal_error "Undefined zone ($client)" unless $clientwild || defined_zone( $client );
+
+    my $serverwild = ( "\L$server" eq 'all' );
+
+    fatal_error "Undefined zone ($server)" unless $serverwild || defined_zone( $server );
+
+    my ( $policy, $default, $remainder ) = split( /:/, $originalpolicy, 3 );
+
+    fatal_error "Invalid or missing POLICY ($originalpolicy)" unless $policy;
+
+    fatal_error "Invalid default action ($default:$remainder)" if defined $remainder;
+
+    ( $policy , my $queue ) = get_target_param $policy;
+    
+    if ( $default ) {
+	if ( "\L$default" eq 'none' ) {
+	    $default = 'none';
+	} elsif ( $actions{$default} ) {
+	    use_policy_action( $default );
+	} else {
+	    fatal_error "Unknown Default Action ($default)";
+	}
+    } else {
+	$default = $default_actions{$policy} || '';
+    }
+
+    fatal_error "Invalid policy ($policy)" unless exists $validpolicies{$policy};
+
+    if ( defined $queue ) {
+	fatal_error "Invalid policy ($policy($queue))" unless $policy eq 'NFQUEUE';
+	require_capability( 'NFQUEUE_TARGET', 'An NFQUEUE Policy', 's' );
+	my $queuenum = numeric_value( $queue );
+	fatal_error "Invalid NFQUEUE queue number ($queue)" unless defined( $queuenum) && $queuenum <= 65535;
+	$policy = "NFQUEUE --queue-num $queuenum";
+    } elsif ( $policy eq 'NONE' ) {
+	fatal_error "NONE policy not allowed with \"all\""
+	    if $clientwild || $serverwild;
+	fatal_error "NONE policy not allowed to/from firewall zone"
+	    if ( zone_type( $client ) == FIREWALL ) || ( zone_type( $server ) == FIREWALL );
+    }
+
+    unless ( $clientwild || $serverwild ) {
+	if ( zone_type( $server ) == BPORT ) {
+	    fatal_error "Invalid policy - DEST zone is a Bridge Port zone but the SOURCE zone is not associated with the same bridge"
+		unless find_zone( $client )->{bridge} eq find_zone( $server)->{bridge} || single_interface( $client ) eq find_zone( $server )->{bridge};
+	}
+    }
+
+    my $chain = rules_chain( ${client}, ${server} );
+    my $chainref;
+
+    if ( defined $filter_table->{$chain} ) {
+	$chainref = $filter_table->{$chain};
+
+	if ( $chainref->{is_policy} ) {
+	    if ( $chainref->{provisional} ) {
+		$chainref->{provisional} = 0;
+		$chainref->{policy} = $policy;
+	    } else {
+		fatal_error qq(Policy "$client $server $policy" duplicates earlier policy "@{$chainref->{policypair}} $chainref->{policy}");
+	    }
+	} elsif ( $chainref->{policy} ) {
+	    fatal_error qq(Policy "$client $server $policy" duplicates earlier policy "@{$chainref->{policypair}} $chainref->{policy}");
+	} else {
+	    convert_to_policy_chain( $chainref, $client, $server, $policy, 0 );
+	    push @policy_chains, ( $chainref ) unless $config{EXPAND_POLICIES} && ( $clientwild || $serverwild );
+	}
+    } else {
+	$chainref = new_policy_chain $client, $server, $policy, 0;
+	push @policy_chains, ( $chainref ) unless $config{EXPAND_POLICIES} && ( $clientwild || $serverwild );
+    }
+
+    $chainref->{loglevel}  = validate_level( $loglevel ) if defined $loglevel && $loglevel ne '';
+
+    if ( $synparams ne '' || $connlimit ne '' ) {
+	my $value = '';
+	fatal_error "Invalid CONNLIMIT ($connlimit)" if $connlimit =~ /^!/;
+	$value  = do_ratelimit $synparams, 'ACCEPT'  if $synparams ne '';
+	$value .= do_connlimit $connlimit            if $connlimit ne '';
+	$chainref->{synparams} = $value;
+	$chainref->{synchain}  = $chain
+    }
+
+    $chainref->{default} = $default if $default;
+
+    if ( $clientwild ) {
+	if ( $serverwild ) {
+	    for my $zone ( @zonelist ) {
+		for my $zone1 ( @zonelist ) {
+		    set_policy_chain $client, $server, rules_chain( ${zone}, ${zone1} ), $chainref, $policy;
+		    print_policy $zone, $zone1, $policy, $chain;
+		}
+	    }
+	} else {
+	    for my $zone ( all_zones ) {
+		set_policy_chain $client, $server, rules_chain( ${zone}, ${server} ), $chainref, $policy;
+		print_policy $zone, $server, $policy, $chain;
+	    }
+	}
+    } elsif ( $serverwild ) {
+	for my $zone ( @zonelist ) {
+	    set_policy_chain $client, $server, rules_chain( ${client}, ${zone} ), $chainref, $policy;
+	    print_policy $client, $zone, $policy, $chain;
+	}
+
+    } else {
+	print_policy $client, $server, $policy, $chain;
+    }
+}
+
+#
+# Generate contents of the /var/lib/shorewall[6]/.policies file as 'here documents' in the generated script
+#
+sub save_policies() {
+    for my $zone1 ( all_zones ) {
+	for my $zone2 ( all_zones ) {
+	    my $chainref  = $filter_table->{ rules_chain( $zone1, $zone2 ) };
+	    my $policyref = $filter_table->{ $chainref->{policychain} };
+
+	    if ( $policyref->{referenced} ) {
+		emit_unindented "$zone1 \t=>\t$zone2\t" . $policyref->{policy} . ' using chain ' . $policyref->{name};
+	    } elsif ( $zone1 ne $zone2 ) {
+		emit_unindented "$zone1 \t=>\t$zone2\t" . $policyref->{policy};
+	    }
+	}
+    }
+}
+
+#
+# Process the policy file
+#
+sub process_policies()
+{
+    our %validpolicies = (
+			  ACCEPT => undef,
+			  REJECT => undef,
+			  DROP   => undef,
+			  CONTINUE => undef,
+			  QUEUE => undef,
+			  NFQUEUE => undef,
+			  NONE => undef
+			  );
+
+    our %map = ( DROP_DEFAULT    => 'DROP' ,
+		 REJECT_DEFAULT  => 'REJECT' ,
+		 ACCEPT_DEFAULT  => 'ACCEPT' ,
+		 QUEUE_DEFAULT   => 'QUEUE' ,
+		 NFQUEUE_DEFAULT => 'NFQUEUE' );
+
+    my $zone;
+    my $firewall = firewall_zone;
+    our @zonelist = $config{EXPAND_POLICIES} ? all_zones : ( all_zones, 'all' );
+
+    for my $option qw( DROP_DEFAULT REJECT_DEFAULT ACCEPT_DEFAULT QUEUE_DEFAULT NFQUEUE_DEFAULT) {
+	my $action = $config{$option};
+	next if $action eq 'none';
+	my $actiontype = $targets{$action};
+												 
+	if ( defined $actiontype ) {
+	    fatal_error "Invalid setting ($action) for $option" unless $actiontype & ACTION;
+	} else {
+	    fatal_error "Default Action $option=$action not found";
+	}
+
+	use_policy_action( $action );
+
+	$default_actions{$map{$option}} = $action;
+    }
+
+    for $zone ( all_zones ) {
+	push @policy_chains, ( new_policy_chain $zone,         $zone, 'ACCEPT', PROVISIONAL );
+	push @policy_chains, ( new_policy_chain firewall_zone, $zone, 'NONE',   PROVISIONAL ) if zone_type( $zone ) == BPORT;
+
+	my $zoneref = find_zone( $zone );
+
+	if ( $config{IMPLICIT_CONTINUE} && ( @{$zoneref->{parents}} || $zoneref->{type} == VSERVER ) ) {
+	    for my $zone1 ( all_zones ) {
+		unless( $zone eq $zone1 ) {
+		    add_or_modify_policy_chain( $zone, $zone1 );
+		    add_or_modify_policy_chain( $zone1, $zone );
+		}
+	    }
+	}
+    }
+
+    if ( my $fn = open_file 'policy' ) {
+	first_entry "$doing $fn...";
+	process_a_policy while read_a_line;
+    } else {
+	fatal_error q(The 'policy' file does not exist or has zero size);
+    }
+
+    for $zone ( all_zones ) {
+	for my $zone1 ( all_zones ) {
+	    fatal_error "No policy defined from zone $zone to zone $zone1" unless $filter_table->{rules_chain( ${zone}, ${zone1} )}{policy};
+	}
+    }
+}
+
+#
+# Policy Rule application
+#
+sub policy_rules( $$$$$ ) {
+    my ( $chainref , $target, $loglevel, $default, $dropmulticast ) = @_;
+
+    unless ( $target eq 'NONE' ) {
+	add_rule $chainref, "-d 224.0.0.0/4 -j RETURN" if $dropmulticast && $target ne 'CONTINUE' && $target ne 'ACCEPT';
+	add_jump $chainref, $default, 0 if $default && $default ne 'none';
+	log_rule $loglevel , $chainref , $target , '' if $loglevel ne '';
+	fatal_error "Null target in policy_rules()" unless $target;
+
+	add_jump( $chainref , $target eq 'REJECT' ? 'reject' : $target, 1 ) unless $target eq 'CONTINUE';
+    }
+}
+
+sub report_syn_flood_protection() {
+    progress_message_nocompress '      Enabled SYN flood protection';
+}
+
+#
+# Complete a policy chain - Add policy-enforcing rules and syn flood, if specified
+#
+sub default_policy( $$$ ) {
+    my $chainref   = $_[0];
+    my $policyref  = $filter_table->{$chainref->{policychain}};
+    my $synparams  = $policyref->{synparams};
+    my $default    = $policyref->{default};
+    my $policy     = $policyref->{policy};
+    my $loglevel   = $policyref->{loglevel};
+
+    assert( $policyref );
+
+    if ( $chainref eq $policyref ) {
+	policy_rules $chainref , $policy, $loglevel , $default, $config{MULTICAST};
+    } else {
+	if ( $policy eq 'ACCEPT' || $policy eq 'QUEUE' || $policy =~ /^NFQUEUE/ ) {
+	    if ( $synparams ) {
+		report_syn_flood_protection;
+		policy_rules $chainref , $policy , $loglevel , $default, $config{MULTICAST};
+	    } else {
+		add_jump $chainref,  $policyref, 1;
+		$chainref = $policyref;
+	    }
+	} elsif ( $policy eq 'CONTINUE' ) {
+	    report_syn_flood_protection if $synparams;
+	    policy_rules $chainref , $policy , $loglevel , $default, $config{MULTICAST};
+	} else {
+	    report_syn_flood_protection if $synparams;
+	    add_jump $chainref , $policyref, 1;
+	    $chainref = $policyref;
+	}
+    }
+
+    progress_message_nocompress "   Policy $policy from $_[1] to $_[2] using chain $chainref->{name}";
+
+}
+
+sub ensure_rules_chain( $ );
+
+#
+# Finish all policy Chains
+#
+sub apply_policy_rules() {
+    progress_message2 'Applying Policies...';
+
+    for my $chainref ( @policy_chains ) {
+	my $policy      = $chainref->{policy};
+
+	unless ( $policy eq 'NONE' ) {
+	    my $loglevel    = $chainref->{loglevel};
+	    my $provisional = $chainref->{provisional};
+	    my $default     = $chainref->{default};
+	    my $name        = $chainref->{name};
+	    my $synparms    = $chainref->{synparms};
+
+	    unless ( $chainref->{referenced} || $provisional || $policy eq 'CONTINUE' ) {
+		if ( $config{OPTIMIZE} & 2 ) {
+		    #
+		    # This policy chain is empty and the only thing that we would put in it is
+		    # the policy-related stuff. Don't create it if all we are going to put in it
+		    # is a single jump. Generate_matrix() will just use the policy target when
+		    # needed.
+		    #
+		    ensure_rules_chain $name if $default ne 'none' || $loglevel || $synparms || $config{MULTICAST} || ! ( $policy eq 'ACCEPT' || $config{FASTACCEPT} );
+		} else {
+		    ensure_rules_chain $name;
+		}
+	    }
+
+	    if ( $name =~ /^all[-2]|[-2]all$/ ) {
+		run_user_exit $chainref;
+		policy_rules $chainref , $policy, $loglevel , $default, $config{MULTICAST};
+	    }
+	}
+    }
+
+    for my $zone ( all_zones ) {
+	for my $zone1 ( all_zones ) {
+	    my $chainref = $filter_table->{rules_chain( ${zone}, ${zone1} )};
+
+	    if ( $chainref->{referenced} ) {
+		run_user_exit $chainref;
+		default_policy $chainref, $zone, $zone1;
+	    }
+	}
+    }
+}
+
+#
+# Complete a standard chain
+#
+#	- run any supplied user exit
+#	- search the policy file for an applicable policy and add rules as
+#	  appropriate
+#	- If no applicable policy is found, add rules for an assummed
+#	  policy of DROP INFO
+#
+sub complete_standard_chain ( $$$$ ) {
+    my ( $stdchainref, $zone, $zone2, $default ) = @_;
+
+    add_rule $stdchainref, "$globals{STATEMATCH} ESTABLISHED,RELATED -j ACCEPT" unless $config{FASTACCEPT};
+
+    run_user_exit $stdchainref;
+
+    my $ruleschainref = $filter_table->{rules_chain( ${zone}, ${zone2} ) } || $filter_table->{rules_chain( 'all', 'all' ) };
+    my ( $policy, $loglevel, $defaultaction ) = ( $default , 6, $config{$default . '_DEFAULT'} );
+    my $policychainref;
+
+    $policychainref = $filter_table->{$ruleschainref->{policychain}} if $ruleschainref;
+
+    ( $policy, $loglevel, $defaultaction ) = @{$policychainref}{'policy', 'loglevel', 'default' } if $policychainref;
+
+    policy_rules $stdchainref , $policy , $loglevel, $defaultaction, 0;
+}
+
+#
+# Create and populate the synflood chains corresponding to entries in /etc/shorewall/policy
+#
+sub setup_syn_flood_chains() {
+    my @zones = ( non_firewall_zones );
+    for my $chainref ( @policy_chains ) {
+	my $limit = $chainref->{synparams};
+	if ( $limit && ! $filter_table->{syn_flood_chain $chainref} ) {
+	    my $level = $chainref->{loglevel};
+	    my $synchainref = @zones > 1 ? 
+		    new_chain 'filter' , syn_flood_chain $chainref :
+		    new_chain( 'filter' , '@' . $chainref->{name} );
+	    add_rule $synchainref , "${limit}-j RETURN";
+	    log_rule_limit( $level ,
+			    $synchainref ,
+			    $chainref->{name} ,
+			    'DROP',
+			    $globals{LOGLIMIT} || '-m limit --limit 5/min --limit-burst 5 ' ,
+			    '' ,
+			    'add' ,
+			    '' )
+		if $level ne '';
+	    add_rule $synchainref, '-j DROP';
+	}
+    }
+}
+
+#
+# Optimize Policy chains with ACCEPT policy
+#
+sub optimize_policy_chains() {
+    for my $chainref ( grep $_->{policy} eq 'ACCEPT', @policy_chains ) {
+	optimize_chain ( $chainref );
+    }
+    #
+    # Often, fw->all has an ACCEPT policy. This code allows optimization in that case
+    #
+    my $outputrules = $filter_table->{OUTPUT}{rules};
+
+    if ( @{$outputrules} && $outputrules->[-1] =~ /-j ACCEPT/ ) {
+	optimize_chain( $filter_table->{OUTPUT} );
+    }
+
+    progress_message '  Policy chains optimized';
+    progress_message '';
+}
+
+################################################################################
+# Modules moved from the Chains module in 4.4.18
+################################################################################
+
+sub finish_chain_section( $$ );
+
+#
+# Create a rules chain if necessary and populate it with the appropriate ESTABLISHED,RELATED rule(s) and perform SYN rate limiting.
+#
+# Return a reference to the chain's table entry.
+#
+sub ensure_rules_chain( $ )
+{
+    my ($chain) = @_;
+
+    my $chainref = ensure_chain 'filter', $chain;
+
+    unless ( $chainref->{referenced} ) {
+	if ( $section eq 'NEW' or $section eq 'DONE' ) {
+	    finish_chain_section $chainref , 'ESTABLISHED,RELATED';
+	} elsif ( $section eq 'RELATED' ) {
+	    finish_chain_section $chainref , 'ESTABLISHED';
+	}
+
+	$chainref->{referenced} = 1;
+    }
+
+    $chainref;
+}
+
+#
+# Add ESTABLISHED,RELATED rules and synparam jumps to the passed chain
+#
+sub finish_chain_section ($$) {
+    my ($chainref, $state ) = @_;
+    my $chain = $chainref->{name};
+    
+    push_comment(''); #These rules should not have comments
+
+    add_rule $chainref, "$globals{STATEMATCH} $state -j ACCEPT" unless $config{FASTACCEPT};
+
+    if ($sections{NEW} ) {
+	if ( $chainref->{is_policy} ) {
+	    if ( $chainref->{synparams} ) {
+		my $synchainref = ensure_chain 'filter', syn_flood_chain $chainref;
+		if ( $section eq 'DONE' ) {
+		    if ( $chainref->{policy} =~ /^(ACCEPT|CONTINUE|QUEUE|NFQUEUE)/ ) {
+			add_jump $chainref, $synchainref, 0, "-p tcp --syn ";
+		    }
+		} else {
+		    add_jump $chainref, $synchainref, 0, "-p tcp --syn ";
+		}
+	    }
+	} else {
+	    my $policychainref = $filter_table->{$chainref->{policychain}};
+	    if ( $policychainref->{synparams} ) {
+		my $synchainref = ensure_chain 'filter', syn_flood_chain $policychainref;
+		add_jump $chainref, $synchainref, 0, "-p tcp --syn ";
+	    }
+	}
+
+	$chainref->{new} = @{$chainref->{rules}};
+    }
+
+    pop_comment;
+}
+
+#
+# Do section-end processing
+#
+sub finish_section ( $ ) {
+    my $sections = $_[0];
+
+    $sections{$_} = 1 for split /,/, $sections;
+
+    for my $zone ( all_zones ) {
+	for my $zone1 ( all_zones ) {
+	    my $chainref = $chain_table{'filter'}{rules_chain( $zone, $zone1 )};
+	    finish_chain_section $chainref, $sections if $chainref->{referenced};
+	}
+    }
+}
+################################################################################
+# Functions moved from the Actions module in 4.4.16
+################################################################################
 #
 # Return ( action, level[:tag] ) from passed full action
 #
@@ -128,19 +798,6 @@ sub split_action ( $ ) {
     ( $target, join ":", @a );
 }
 
-#
-# Split the passed target into the basic target and parameter
-#
-sub get_target_param( $ ) {
-    my ( $target, $param ) = split '/', $_[0];
-
-    unless ( defined $param ) {
-	( $target, $param ) = ( $1, $2 ) if $target =~ /^(.*?)[(](.*)[)]$/;
-    }
-
-    ( $target, $param );
-}
-
 #
 # Create a normalized action name from the passed pieces.
 #
@@ -192,6 +849,8 @@ sub new_action( $$ ) {
 
     my ( $action , $type ) = @_;
 
+    fatal_error "Invalid action name($action)" if reserved_name( $action );
+
     $actions{$action} = { actchain => ''  };
 
     $targets{$action} = $type;
@@ -761,13 +1420,15 @@ sub process_action( $) {
 sub process_actions2 () {
     progress_message2 "$doing policy actions...";
 
-    for ( map normalize_action_name $_, ( grep ! ( $targets{$_} & BUILTIN ), policy_actions ) ) {
+    for ( map normalize_action_name $_, ( grep ! ( $targets{$_} & BUILTIN ), keys %policy_actions ) ) {
 	if ( my $ref = use_action( $_ ) ) {
 	    process_action( $ref );
 	}
     }
 }
-
+################################################################################
+# End of functions moved from the Actions module in 4.4.16
+################################################################################
 #
 # Expand a macro rule from the rules file
 #
@@ -1090,6 +1751,7 @@ sub process_rule1 ( $$$$$$$$$$$$$$$$ ) {
 
 	fatal_error "Missing source zone" if $sourcezone eq '-' || $sourcezone =~ /^:/;
 	fatal_error "Unknown source zone ($sourcezone)" unless $sourceref = defined_zone( $sourcezone );
+	fatal_error 'USER/GROUP may only be specified when the SOURCE zone is $FW' unless $user eq '-' || $sourcezone eq firewall_zone;
     }
 
     if ( $actiontype & NATONLY ) {
@@ -1172,7 +1834,7 @@ sub process_rule1 ( $$$$$$$$$$$$$$$$ ) {
 	    #
 	    # Mark the chain as referenced and add appropriate rules from earlier sections.
 	    #
-	    $chainref = ensure_filter_chain $chain, 1;
+	    $chainref = ensure_rules_chain $chain;
 	    #
 	    # Don't let the rules in this chain be moved elsewhere
 	    #
@@ -1352,6 +2014,7 @@ sub process_rule1 ( $$$$$$$$$$$$$$$$ ) {
 	    $loglevel = '';
 	    $dest     = $server;
 	    $action   = 'ACCEPT';
+	    $origdest = ALLIP if  $origdest =~ /[+]/;
 	}
     } elsif ( $actiontype & NONAT ) {
 	#
@@ -1488,7 +2151,7 @@ sub process_rule1 ( $$$$$$$$$$$$$$$$ ) {
 sub process_section ($) {
     my $sect = shift;
     #
-    # read_a_line has already verified that there are exactly two tokens on the line
+    # split_line1 has already verified that there are exactly two tokens on the line
     #
     fatal_error "Invalid SECTION ($sect)" unless defined $sections{$sect};
     fatal_error "Duplicate or out of order SECTION $sect" if $sections{$sect};
@@ -1574,7 +2237,7 @@ sub build_zone_list( $$$\$\$ ) {
 # Process a Record in the rules file
 #
 sub process_rule ( ) {
-    my ( $target, $source, $dest, $proto, $ports, $sports, $origdest, $ratelimit, $user, $mark, $connlimit, $time, $headers ) = split_line1 1, 13, 'rules file', $rule_commands;
+    my ( $target, $source, $dest, $protos, $ports, $sports, $origdest, $ratelimit, $user, $mark, $connlimit, $time, $headers ) = split_line1 1, 13, 'rules file', $rule_commands;
 
     process_comment,            return 1 if $target eq 'COMMENT';
     process_section( $source ), return 1 if $target eq 'SECTION';
@@ -1596,32 +2259,39 @@ sub process_rule ( ) {
     my $fw        = firewall_zone;
     my @source    = build_zone_list ( $fw, $source, 'SOURCE', $intrazone, $wild );
     my @dest      = build_zone_list ( $fw, $dest,   'DEST'  , $intrazone, $wild );
+    my @protos    = split_list1 $protos, 'Protocol';
     my $generated = 0;
 
     fatal_error "Invalid or missing ACTION ($target)" unless defined $action;
 
+    if ( @protos > 1 ) {
+	fatal_error "Inversion not allowed in a PROTO list" if $protos =~ tr/!/!/;
+    }
+
     for $source ( @source ) {
 	for $dest ( @dest ) {
 	    my $sourcezone = (split( /:/, $source, 2 ) )[0];
 	    my $destzone   = (split( /:/, $dest,   2 ) )[0];
 	    $destzone = $action =~ /^REDIRECT/ ? $fw : '' unless defined_zone $destzone;
 	    if ( ! $wild || $intrazone || ( $sourcezone ne $destzone ) ) {
-		$generated |= process_rule1( undef,
-					     $target,
-					     '',
-					     $source,
-					     $dest,
-					     $proto,
-					     $ports,
-					     $sports,
-					     $origdest,
-					     $ratelimit,
-					     $user,
-					     $mark,
-					     $connlimit,
-					     $time,
-					     $headers,
-					     $wild );
+		for my $proto ( @protos ) {
+		    $generated |= process_rule1( undef,
+						 $target,
+						 '',
+						 $source,
+						 $dest,
+						 $proto,
+						 $ports,
+						 $sports,
+						 $origdest,
+						 $ratelimit,
+						 $user,
+						 $mark,
+						 $connlimit,
+						 $time,
+						 $headers,
+						 $wild );
+		}
 	    }
 	}
     }
