@@ -40,7 +40,7 @@ use strict;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( setup_tc );
 our @EXPORT_OK = qw( process_tc_rule initialize );
-our $VERSION = '4.4_17';
+our $VERSION = '4.4_19';
 
 our %tcs = ( T => { chain  => 'tcpost',
 		    connmark => 0,
@@ -252,10 +252,18 @@ sub process_tc_rule( ) {
 	    require_capability ('CONNMARK' , "CONNMARK Rules", '' ) if $connmark;
 
 	} else {
-	    fatal_error "Invalid MARK ($originalmark)"   unless $mark =~ /^([0-9]+|0x[0-9a-f]+)$/ and $designator =~ /^([0-9]+|0x[0-9a-f]+)$/;
+	    fatal_error "Invalid MARK ($originalmark)"   unless $mark =~ /^([0-9a-fA-F]+)$/ and $designator =~ /^([0-9a-fA-F]+)$/;
 
 	    if ( $config{TC_ENABLED} eq 'Internal' || $config{TC_ENABLED} eq 'Shared' ) {
+		$originalmark = join( ':', normalize_hex( $mark ), normalize_hex( $designator ) );
 		fatal_error "Unknown Class ($originalmark)}" unless ( $device = $classids{$originalmark} );
+		fatal_error "IFB Classes may not be specified in tcrules" if @{$tcdevices{$device}{redirected}};
+
+		if ( $dest eq '-' ) {
+		    $dest = $device;
+		} else {
+		    $dest = join( ':', $device, $dest ) unless $dest =~ /^[[:alpha:]]/;
+		}
 	    }
 
 	    $chain   = 'tcpost';
@@ -404,6 +412,8 @@ sub process_tc_rule( ) {
 	}
     }
 
+    fatal_error "USER/GROUP only allowed in the OUTPUT chain" unless ( $user eq '-' || ( $chain eq 'tcout' || $chain eq 'tcpost' ) ); 
+
     if ( ( my $result = expand_rule( ensure_chain( 'mangle' , $chain ) ,
 				     $restrictions{$chain} | $restriction,
 				     do_proto( $proto, $ports, $sports) .
@@ -517,7 +527,8 @@ sub process_simple_device() {
 	 );
 
     emit ( "run_tc qdisc add dev $physical handle ffff: ingress",
-	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n"
+	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip src "  . ALLIPv4 . " police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n",
+	   "run_tc filter add dev $physical parent ffff: protocol all prio 10 u32 match ip6 src " . ALLIPv6 . " police rate ${in_bandwidth}kbit burst $in_burst drop flowid :1\n"
 	 ) if $in_bandwidth;
 
     if ( $out_part ne '-' ) {
@@ -566,10 +577,13 @@ sub process_simple_device() {
 
     for ( my $i = 1; $i <= 3; $i++ ) {
 	emit "run_tc qdisc add dev $physical parent $number:$i handle ${number}${i}: sfq quantum 1875 limit 127 perturb 10";
-	emit "run_tc filter add dev $physical protocol all parent $number: handle $i fw classid $number:$i";
+	emit "run_tc filter add dev $physical protocol all prio 2 parent $number: handle $i fw classid $number:$i";
 	emit "run_tc filter add dev $physical protocol all prio 1 parent ${number}$i: handle ${number}${i} flow hash keys $type divisor 1024" if $type ne '-' && have_capability 'FLOW_FILTER';
 	emit '';
     }
+    
+    emit "run_tc filter add dev $physical parent $number:0 protocol all prio 1 u32 match ip protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $number:1\n";
+    emit "run_tc filter add dev $physical parent $number:0 protocol all prio 1 u32 match ip6 protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $number:1\n";
 
     save_progress_message_short qq("   TC Device $physical defined.");
 
@@ -598,15 +612,16 @@ sub validate_tc_device( ) {
 	fatal_error "Invalid NUMBER:INTERFACE ($device:$number:$rest)" if defined $rest;
 
 	if ( defined $number ) {
+	    $number = normalize_hex( $number );
 	    $devnumber = hex_value( $number );
-	    fatal_error "Invalid interface NUMBER ($number)" unless defined $devnumber && $devnumber;
+	    fatal_error "Invalid device NUMBER ($number)" unless defined $devnumber && $devnumber && $devnumber < 256;
 	    fatal_error "Duplicate interface number ($number)" if defined $devnums[ $devnumber ];
 	    $devnum = $devnumber if $devnumber > $devnum;
 	} else {
 	    fatal_error "Missing interface NUMBER";
 	}
-    } else {
-	$devnumber = ++$devnum;
+    } elsif ( ( $devnumber = ++$devnum ) > 255 ) {
+	fatal_error "Attempting to assign a device number > 255";
     }
 
     $devnums[ $devnumber ] = $device;
@@ -648,11 +663,26 @@ sub validate_tc_device( ) {
 	    fatal_error "Invalid device name ($rdevice)" if $rdevice =~ /[:+]/;
 	    my $rdevref = $tcdevices{$rdevice};
 	    fatal_error "REDIRECTED device ($rdevice) has not been defined in this file" unless $rdevref;
-	    fatal_error "IN-BANDWIDTH must be zero for REDIRECTED devices" if $rdevref->{in_bandwidth} ne '0kbit';
+	    fatal_error "IN-BANDWIDTH must be zero for REDIRECTED devices" if $rdevref->{in_bandwidth} != 0;
 	}
     }
 
-    $tcdevices{$device} = { in_bandwidth  => rate_to_kbit( $inband ) . 'kbit',
+    my $in_burst = '10kb';
+
+    if ( $inband =~ /:/ ) {
+	my ( $in_band, $burst ) = split /:/, $inband, 2;
+
+	if ( defined $burst && $burst ne '' ) {
+	    fatal_error "Invalid IN-BANDWIDTH" if $burst =~ /:/;
+	    fatal_error "Invalid burst ($burst)" unless $burst =~ /^\d+(k|kb|m|mb|mbit|kbit|b)?$/;
+	    $in_burst = $burst;
+	}
+
+	$inband = $in_band;
+    }
+
+    $tcdevices{$device} = { in_bandwidth  => rate_to_kbit( $inband ),
+			    in_burst      => $in_burst,
 			    out_bandwidth => rate_to_kbit( $outband ) . 'kbit',
 			    number        => $devnumber,
 			    classify      => $classify,
@@ -726,7 +756,6 @@ sub dev_by_number( $ ) {
     }
 
     ( $dev , $devref );
-
 }
 
 sub validate_tc_class( ) {
@@ -742,7 +771,7 @@ sub validate_tc_class( ) {
 	( $device, my ($number, $subnumber, $rest ) )  = split /:/, $device, 4;
 	fatal_error "Invalid INTERFACE:CLASS ($devclass)" if defined $rest;
 
-	if ( $device =~ /^(\d+|0x[\da-fA-F]+)$/ ) {
+	if ( $device =~ /^[\da-fA-F]+$/ && ! $tcdevices{$device} ) {
 	    ( $number , $classnumber ) = ( hex_value $device, hex_value $number );
 	    ( $device , $devref) = dev_by_number( $number );
 	} else {
@@ -758,8 +787,9 @@ sub validate_tc_class( ) {
 		$classnumber = hex_value $subnumber;
 	    }
 
-	    fatal_error "Invalid interface/class number ($devclass)" unless defined $classnumber && $classnumber;
-	    fatal_error "Duplicate interface/class number ($devclass)" if defined $devnums[ $classnumber ];
+	    fatal_error "Invalid interface/class number ($devclass)" unless defined $classnumber && $classnumber && $classnumber < 0x8000;
+	    fatal_error "Reserved class number (1)" if $classnumber == 1;
+	    fatal_error "Duplicate interface:class number ($number:$classnumber}" if $tcclasses{$device}{$classnumber};
 	} else {
 	    fatal_error "Missing interface NUMBER";
 	}
@@ -951,13 +981,21 @@ sub process_tc_filter() {
 
     my ($device, $class, $rest ) = split /:/, $devclass, 3;
 
+    our $lastdevice;
+
     fatal_error "Invalid INTERFACE:CLASS ($devclass)" if defined $rest || ! ($device && $class );
 
     my ( $ip, $ip32, $prio , $lo ) = $family == F_IPV4 ? ('ip', 'ip', 10, 2 ) : ('ipv6', 'ip6', 11 , 4 );
 
-    ( $device , my $devref ) = dev_by_number( $device );
+    my $devref;
 
-    my $devnum = $devref->{number};
+    if ( $device =~ /^[\da-fA-F]+$/ && ! $tcdevices{$device} ) {
+	( $device, $devref ) = dev_by_number( hex_value( $device ) );
+    } else {
+	( $device , $devref ) = dev_by_number( $device );
+    }
+
+    my $devnum = in_hexp $devref->{number};
 
     my $tcref = $tcclasses{$device};
 
@@ -972,6 +1010,17 @@ sub process_tc_filter() {
     fatal_error "Unknown CLASS ($devclass)"                  unless $tcref && $tcref->{occurs};
     fatal_error "Filters may not specify an occurring CLASS" if $tcref->{occurs} > 1;
 
+    if ( $devref->{physical} ne $lastdevice ) {
+	if ( $lastdevice ) {
+	    pop_indent;
+	    emit "fi\n";
+	}
+
+	$lastdevice = $devref->{physical};
+	emit "if interface_is_up $lastdevice; then";
+	push_indent;
+    }
+
     my $rule = "filter add dev $devref->{physical} protocol $ip parent $devnum:0 prio $prio u32";
 
     if ( $source ne '-' ) {
@@ -1019,7 +1068,7 @@ sub process_tc_filter() {
 
     if ( $portlist eq '-' && $sportlist eq '-' ) {
 	emit( "\nrun_tc $rule\\" ,
-	      "   flowid $devref->{number}:$class" ,
+	      "   flowid $devnum:$class" ,
 	      '' );
     } else {
 	fatal_error "Ports may not be specified without a PROTO" unless $protonumber;
@@ -1081,7 +1130,7 @@ sub process_tc_filter() {
 
 		    emit( "\nrun_tc $rule\\" ,
 			  "   $rule1\\" ,
-			  "   flowid $devref->{number}:$class" );
+			  "   flowid $devnum:$class" );
 		}
 	    }
 	} else {
@@ -1093,24 +1142,24 @@ sub process_tc_filter() {
 		    fatal_error "ICMP not allowed with IPv6" unless $family == F_IPV4;
 		    fatal_error "SOURCE PORT(S) are not allowed with ICMP" if $sportlist ne '-';
 
-		    my ( $icmptype , $icmpcode ) = split '//', validate_icmp( $portrange );
+		    my ( $icmptype , $icmpcode ) = split '/', validate_icmp( $portrange );
 
 		    my $rule1 = "   match icmp type $icmptype 0xff";
 		    $rule1   .= "\\\n   match icmp code $icmpcode 0xff" if defined $icmpcode;
 		    emit( "\nrun_tc ${rule}\\" ,
 			  "$rule1\\" ,
-			  "   flowid $devref->{number}:$class" );
+			  "   flowid $devnum:$class" );
 		} elsif ( $protonumber == IPv6_ICMP ) {
 		    fatal_error "IPv6 ICMP not allowed with IPv4" unless $family == F_IPV4;
 		    fatal_error "SOURCE PORT(S) are not allowed with IPv6 ICMP" if $sportlist ne '-';
 
-		    my ( $icmptype , $icmpcode ) = split '//', validate_icmp6( $portrange );
+		    my ( $icmptype , $icmpcode ) = split '/', validate_icmp6( $portrange );
 
 		    my $rule1 = "   match icmp6 type $icmptype 0xff";
 		    $rule1   .= "\\\n   match icmp6 code $icmpcode 0xff" if defined $icmpcode;
 		    emit( "\nrun_tc ${rule}\\" ,
 			  "$rule1\\" ,
-			  "   flowid $devref->{number}:$class" );
+			  "   flowid $devnum:$class" );
 		} else {
 		    my @portlist = expand_port_range $protonumber , $portrange;
 
@@ -1130,7 +1179,7 @@ sub process_tc_filter() {
 			if ( $sportlist eq '-' ) {
 			    emit( "\nrun_tc ${rule}\\" ,
 				  "   $rule1\\" ,
-				  "   flowid $devref->{number}:$class" );
+				  "   flowid $devnum:$class" );
 			} else {
 			    for my $sportrange ( split_list $sportlist , 'port list' ) {
 				my @sportlist = expand_port_range $protonumber , $sportrange;
@@ -1151,7 +1200,7 @@ sub process_tc_filter() {
 				    emit( "\nrun_tc ${rule}\\",
 					  "   $rule1\\" ,
 					  "   $rule2\\" ,
-					  "   flowid $devref->{number}:$class" );
+					  "   flowid $devnum:$class" );
 				}
 			    }
 			}
@@ -1186,6 +1235,8 @@ sub process_tcfilters() {
 
     my $fn = open_file 'tcfilters';
 
+    our $lastdevice = '';
+
     if ( $fn ) {
 	my @family = ( $family );
 	
@@ -1213,6 +1264,12 @@ sub process_tcfilters() {
 	}
 
 	Shorewall::IPAddrs::initialize( $family = pop @family );
+
+	if ( $lastdevice ) {
+	    pop_indent;
+	    emit "fi\n";
+	}
+
     }
 }
 
@@ -1290,7 +1347,7 @@ sub setup_simple_traffic_shaping() {
 	first_entry
 	    sub {
 		progress_message2 "$doing $fn1...";
-		warning_message "There are entries in $fn1 but $fn was empty" unless $interfaces;
+		warning_message "There are entries in $fn1 but $fn was empty" unless $interfaces || $family == F_IPV6;
 	    };
 
 	process_tc_priority while read_a_line;
@@ -1336,6 +1393,8 @@ sub setup_traffic_shaping() {
 	my $devnum  = in_hexp $devref->{number};
 	my $r2q     = int calculate_r2q $devref->{out_bandwidth};
 
+	fatal_error "No default class defined for device $device" unless $devref->{default};
+
 	$device = physical_name $device;
 
 	my $dev = chain_base( $device );
@@ -1378,26 +1437,9 @@ sub setup_traffic_shaping() {
 		      qq(fi) );
 	    }
 
-	    my $in_burst = '10kb';
-	    my $inband;
-
-	    if ( $devref->{in_bandwidth} =~ /:/ ) {
-		my ( $in_band, $burst ) = split /:/, $devref->{in_bandwidth}, 2;
-
-		if ( defined $burst && $burst ne '' ) {
-		    fatal_error "Invalid IN-BANDWIDTH" if $burst =~ /:/;
-		    fatal_error "Invalid burst ($burst)" unless $burst =~ /^\d+(k|kb|m|mb|mbit|kbit|b)?$/;
-		    $in_burst = $burst;
-		}
-
-		$inband = rate_to_kbit( $in_band );
-	    } else {
-		$inband = rate_to_kbit $devref->{in_bandwidth};
-	    }
-
-	    if ( $inband ) {
+	    if ( $devref->{in_bandwidth} ) {
 		emit ( "run_tc qdisc add dev $device handle ffff: ingress",
-		       "run_tc filter add dev $device parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate ${inband}kbit burst $in_burst drop flowid :1"
+		       "run_tc filter add dev $device parent ffff: protocol all prio 10 u32 match ip src 0.0.0.0/0 police rate $devref->{in_bandwidth}kbit burst $devref->{in_burst} drop flowid :1"
 		     );
 	    }
 
@@ -1490,7 +1532,7 @@ sub setup_traffic_shaping() {
 	    #
 	    # options
 	    #
-	    emit "run_tc filter add dev $device parent $devref->{number}:0 protocol ip prio " . ( $priority | 10 ) ." u32 match ip protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $classid" if $tcref->{tcp_ack};
+	    emit "run_tc filter add dev $device parent $devicenumber:0 protocol ip prio " . ( $priority | 10 ) ." u32 match ip protocol 6 0xff match u8 0x05 0x0f at 0 match u16 0x0000 0xffc0 at 2 match u8 0x10 0xff at 33 flowid $classid" if $tcref->{tcp_ack};
 
 	    for my $tospair ( @{$tcref->{tos}} ) {
 		my ( $tos, $mask ) = split q(/), $tospair;
