@@ -22,7 +22,7 @@
 #       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 #
 
-VERSION=4.4.17
+VERSION=4.4.19.3
 
 usage() # $1 = exit status
 {
@@ -107,6 +107,9 @@ fi
 
 SPARSE=
 MANDIR=${MANDIR:-"/usr/share/man"}
+[ -n "${LIBEXEC:=share}" ]
+[ -n "${PERLLIB:=share/shorewall}" ]
+
 INSTALLD='-D'
 
 case $(uname) in
@@ -233,9 +236,13 @@ fi
 if [ -z "$CYGWIN" ]; then
    install_file shorewall ${DESTDIR}/sbin/shorewall 0755
    echo "shorewall control program installed in ${DESTDIR}/sbin/shorewall"
+   eval sed -i \'s\|g_libexec=.\*\|g_libexec=$LIBEXEC\|\' ${DESTDIR}/sbin/shorewall
+   eval sed -i \'s\|g_perllib=.\*\|g_perllib=$PERLLIB\|\' ${DESTDIR}/sbin/shorewall
 else
    install_file shorewall ${DESTDIR}/bin/shorewall 0755
    echo "shorewall control program installed in ${DESTDIR}/bin/shorewall"
+   eval sed -i \'s\|g_libexec=.\*\|g_libexec=$LIBEXEC\|\' ${DESTDIR}/bin/shorewall
+   eval sed -i \'s\|g_perllib=.\*\|g_perllib=$PERLLIB\|\' ${DESTDIR}/bin/shorewall
 fi
 
 #
@@ -258,7 +265,8 @@ fi
 # Create /etc/shorewall, /usr/share/shorewall and /var/shorewall if needed
 #
 mkdir -p ${DESTDIR}/etc/shorewall
-mkdir -p ${DESTDIR}/usr/share/shorewall
+mkdir -p ${DESTDIR}/usr/${LIBEXEC}/shorewall
+mkdir -p ${DESTDIR}/usr/${PERLLIB}/Shorewall
 mkdir -p ${DESTDIR}/usr/share/shorewall/configfiles
 mkdir -p ${DESTDIR}/var/lib/shorewall
 
@@ -323,10 +331,10 @@ delete_file ${DESTDIR}/usr/share/shorewall/prog.footer
 # Install wait4ifup
 #
 
-install_file wait4ifup ${DESTDIR}/usr/share/shorewall/wait4ifup 0755
+install_file wait4ifup ${DESTDIR}/usr/${LIBEXEC}/shorewall/wait4ifup 0755
 
 echo
-echo "wait4ifup installed in ${DESTDIR}/usr/share/shorewall/wait4ifup"
+echo "wait4ifup installed in ${DESTDIR}/usr/${LIBEXEC}/shorewall/wait4ifup"
 
 #
 # Install the policy file
@@ -440,11 +448,16 @@ if [ -z "$SPARSE" -a ! -f ${DESTDIR}/etc/shorewall/notrack ]; then
     echo "Notrack file installed as ${DESTDIR}/etc/shorewall/notrack"
 fi
 #
-# Install the Modules file
+# Install the Modules files
 #
 run_install $OWNERSHIP -m 0644 modules ${DESTDIR}/usr/share/shorewall
 echo "Modules file installed as ${DESTDIR}/usr/share/shorewall/modules"
 
+for f in modules.*; do
+    run_install $OWNERSHIP -m 0644 $f ${DESTDIR}/usr/share/shorewall/$f
+    echo "Module file $f installed as ${DESTDIR}/usr/share/shorewall/$f"
+done
+
 #
 # Install the Module Helpers file
 #
@@ -811,14 +824,14 @@ chmod 755 ${DESTDIR}/usr/share/shorewall/Shorewall
 #
 cd Perl
 
-install_file compiler.pl ${DESTDIR}/usr/share/shorewall/compiler.pl 0755
+install_file compiler.pl ${DESTDIR}/usr/${LIBEXEC}/shorewall/compiler.pl 0755
 
 echo
-echo "Compiler installed in ${DESTDIR}/usr/share/shorewall/compiler.pl"
+echo "Compiler installed in ${DESTDIR}/usr/${LIBEXEC}/shorewall/compiler.pl"
 #
 # Install the params file helper
 #
-install_file getparams ${DESTDIR}/usr/share/shorewall/getparams 0755
+install_file getparams ${DESTDIR}/usr/${LIBEXEC}/shorewall/getparams 0755
 
 echo
 echo "Params file helper installed in ${DESTDIR}/usr/share/shorewall/getparams"
@@ -826,8 +839,8 @@ echo "Params file helper installed in ${DESTDIR}/usr/share/shorewall/getparams"
 # Install the libraries
 #
 for f in Shorewall/*.pm ; do
-    install_file $f ${DESTDIR}/usr/share/shorewall/$f 0644
-    echo "Module ${f%.*} installed as ${DESTDIR}/usr/share/shorewall/$f"
+    install_file $f ${DESTDIR}/usr/${PERLLIB}/$f 0644
+    echo "Module ${f%.*} installed as ${DESTDIR}/usr/${PERLLIB}/$f"
 done
 #
 # Install the program skeleton files
@@ -888,6 +901,7 @@ fi
 if [ -z "$DESTDIR" ]; then
     rm -rf /usr/share/shorewall-perl
     rm -rf /usr/share/shorewall-shell
+    [ "$PERLLIB" != share/shorewall ] && rm -rf /usr/share/shorewall/Shorewall
 fi
 
 if [ -z "$DESTDIR" -a -n "$first_install" -a -z "${CYGWIN}${MAC}" ]; then
